package org.steveshipway.blueprints;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.directive.*;
import org.steveshipway.blueprints.lib.Constants;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;

// Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
// Minecraft 1.8
import org.steveshipway.blueprints.util18.*;
import net.minecraft.util.BlockPos;

public class Blueprint {
	private boolean floaty = false;  // float y coordinate before build
	private boolean dry = false;     // float on water
	private boolean recurse = false; // allow recursion
	private int distance = 2;        // not used
	public boolean visible = true;   // hidden or visible blueprint
	private LinkedList steps = null; // build steps
	public String name;              // my name
	public String helptext = "No description available";
	public float auto_chance = 0;
	public Set biomes = new HashSet(); 
	private Logger logger = null;
	
	public Blueprint(String name, String description, Boolean v, LinkedList steps, Logger logger) {
		// create this blueprint
		this.name = name;
		this.steps = steps;
		this.visible = v;
		this.logger = logger;
		if( description != null && description != "" ) {
			this.helptext = description;
		}
		if( steps.isEmpty() ) {
        	logger.warn("["+name+"] Created a blueprint with no steps?");
        	return;
		}
		// iterate over the commands, removing any options and setting the flags
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        String s = d.description();
	        if( s != null ) {
	        	this.helptext = s;
	        	// this.steps.remove(d);
	        	continue;
	        }
	        s = d.option();
	        if( s != null ) {
	        	if( s.contains("float")) {
	        		this.floaty = true;
	        	}
	        	if( s.contains("dry")) {
	        		this.dry = true;
	        		this.floaty = true;
	        	}
	        	if( s.contains("recurs")) { // catch recurse and recursive
	        		this.recurse = true;
	        	}

	        	// this.steps.remove(d);
	        	continue;
	        }
	        int n = d.distance();
	        if( n >= 0 ) {
	        	this.distance = n;
	        	continue;
	        }
	        float f = d.auto_chance();
	        if( f > 0 && f <= 1 ) {
	        	this.auto_chance = f;
	        	this.biomes = ((AutoDirective)d).biomes;
	        	logger.info("Added an auto-generated plan '"+this.name+"', chance "+Math.floor(this.auto_chance*Constants.CHANCE_BASE)+"/"+Constants.CHANCE_BASE);
	        	continue;
	        }	        
	    }
    	//logger.info("Created a blueprint with "+this.steps.size()+" steps");
	}
	
	public boolean isVisible() {
		return visible;
	}
	public boolean setVisible(boolean v) {
		visible = v;
		return visible;
	}

	public String helpText() {
		if( this.helptext != null ) {
			return this.helptext;
		}
		return "No description available";
	}
	public String name() { return this.name; }
	
	// build this blueprint, at the specified coordinates.
	public boolean build(World world, BlockPos location, int r ) {
		// iterate over steps, actioning each in turn
		// may require (gulp) recursion...
		// return true if it was successfully built
		boolean rv = true;
		if( world.isRemote ) {
        	logger.warn("Blueprint build called with remote world!  Should never happen.");
			return false;			
		}
		if( this.steps == null || this.steps.size() == 0 ) {
        	logger.warn("Build of '"+this.name+"' had no steps!  This seems wrong...");
			return false;
		}
    	logger.info("Build of blueprint '"+this.name+"' starting: "+this.steps.size()+" steps");
		
    	// floating option -- go up.down to find surface of world
    	if( this.floaty ) {
    		BlockPos blk = new BlockPos(location);
    		blk = location.down();
    		Block b = Util.getBlock(world,blk);
    		while( blk.getY() > 0 && b == Blocks.air ) {
    			location = location.down();
    			blk = location.down();
    			b = Util.getBlock(world,blk);
    		}
       		while( blk.getY() < 128 && b != Blocks.air && ( this.dry || b != Blocks.water )) {
    			location = location.up(); 
       			blk = location.up();
       			b = Util.getBlock(world,blk);
    		}
        	logger.info("Floating Y coordinate to surface level: "+location.getY());
    	}
    	
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	    	//CheshireCat.logger.debug("Calling step: "+d.asText());
	        rv = d.build(world, location, r, logger);
	        if(!rv) {
	        	logger.error("Build of '"+this.name+"' failed with step: ["+d.asText()+"]");
	        	break; 
	        }
	    }
    	//CheshireCat.logger.info("Build of blueprint '"+this.name+"' ended");
		return rv;
	}
	public String asText() {
		String s = "";
		Iterator i = this.steps.iterator();
		if( visible ) {
			s = "PUBLIC";
		} else {
			s = "PRIVATE";
		}
		s = s + " PLAN " + name + " " + helptext + "\n";

	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        s = s + "  " + d.asText() + "\n";
	    }
	    s = s + "END\n";
		return s;
	}
	public boolean save(String f) {
		boolean rv = true;
		if( f == null ) { return false; }
		Writer writer = null;
		try {
		    writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream(f), "utf-8"));
		    writer.write("# Blueprint for: "+this.name+"\n");
			writer.write(this.asText()+"\n");
		} catch (IOException ex) {
			logger.error("Unable to write file "+f+": "+ex.getMessage());
			rv = false;
		} finally {
		    try {writer.close();} catch (Exception ex) {}
		}
		return rv;
	}
}
